"""Process a subcommand CLI manifest."""

from dataclasses import dataclass


@dataclass
class ManifestSubcommand:
    """Processed subcommand CLI manifest subcommand info."""

    name: str
    short_help: str
    visible_aliases: list[str]
    hidden_aliases: list[str]


@dataclass
class Manifest:
    """Processed subcommand CLI manifest."""

    subcommands: list[ManifestSubcommand]
    tab_completion_supported: bool
    tab_completion_install_option: str | list[str] | None
    tab_completion_show_option: str | list[str] | None


MANIFEST_TYPE_SINGLE = "slices-cli:subcommand-metadata"
MANIFEST_TYPE_MULTIPLE = "slices-cli:multi-subcommand-metadata"
SUPPORTED_VERSIONS = ("1.0",)

SUPPORTED_FIELDS_SINGLE = (
    "version",
    "tab_completion_supported",
    "tab_completion_install_option",
    "tab_completion_show_option",
    "short_help",
    "visible_aliases",
    "hidden_aliases",
)
SUPPORTED_FIELDS_MULTI = (
    "version",
    "tab_completion_supported",
    "tab_completion_install_option",
    "tab_completion_show_option",
    "subcommands",
)
SUPPORTED_SUBCOMMAND_FIELDS = ("name", "short_help", "visible_aliases", "hidden_aliases")
SUPPORTED_FIELDS_EXTRA = ("cached_cli_version",)

REQUIRED_FIELDS_SINGLE = ("short_help",)
REQUIRED_FIELDS_MULTI = ("subcommands",)
REQUIRED_SUBCOMMAND_FIELDS = ("name", "short_help")


class ManifestException(Exception):
    """An exception while processing a subcommand CLI manifest."""


def process_manifest(
    executable_name: str, manifest_dict: dict, *, require_subcommand_with_exe_name: bool = True
) -> Manifest:
    """Process the manifest dictionary into a Manifest object."""
    assert executable_name.startswith("slices-")
    subcommand_basename = str(executable_name[len("slices-") :])

    manifest_dict = dict(manifest_dict)

    # First handle the @type and version fields
    if "@type" not in manifest_dict or manifest_dict["@type"] not in (MANIFEST_TYPE_SINGLE, MANIFEST_TYPE_MULTIPLE):
        raise ManifestException("Invalid or missing manifest @type")
    is_single = manifest_dict["@type"] == MANIFEST_TYPE_SINGLE
    del manifest_dict["@type"]

    if "version" not in manifest_dict:
        raise ManifestException("Missing manifest version")
    version = manifest_dict["version"]
    del manifest_dict["version"]
    if version not in SUPPORTED_VERSIONS:
        raise ManifestException(f"Unsupported manifest version ({version!r} is not in {', '.join(SUPPORTED_VERSIONS)})")

    # Handle the regular fields
    for field in manifest_dict:
        if field not in ((SUPPORTED_FIELDS_SINGLE if is_single else SUPPORTED_FIELDS_MULTI) + SUPPORTED_FIELDS_EXTRA):
            raise ManifestException(f"Unsupported manifest field ({field!r})")
    for field in REQUIRED_FIELDS_SINGLE if is_single else REQUIRED_FIELDS_MULTI:
        if field not in manifest_dict:
            raise ManifestException(f"Missing required manifest field ({field!r})")

    subcommands = []
    if is_single:
        subcommands.append(
            ManifestSubcommand(
                name=subcommand_basename,
                short_help=manifest_dict["short_help"],
                visible_aliases=manifest_dict.get("visible_aliases", []),
                hidden_aliases=manifest_dict.get("hidden_aliases", []),
            )
        )
    else:
        for index, subcommand in enumerate(manifest_dict["subcommands"]):
            for field in subcommand:
                if field not in SUPPORTED_SUBCOMMAND_FIELDS:
                    raise ManifestException(f"Unsupported subcommand field ({field!r}) for subcommand {index}")
            for field in REQUIRED_SUBCOMMAND_FIELDS:
                if field not in subcommand:
                    raise ManifestException(f"Missing required subcommand field ({field!r}) for subcommand {index}")
            if subcommand["name"] in [s.name for s in subcommands]:
                raise ManifestException(f"Duplicate subcommand name {subcommand['name']!r}")
            subcommands.append(
                ManifestSubcommand(
                    name=subcommand["name"],
                    short_help=subcommand["short_help"],
                    visible_aliases=subcommand.get("visible_aliases", []),
                    hidden_aliases=subcommand.get("hidden_aliases", []),
                )
            )
        if require_subcommand_with_exe_name and subcommand_basename not in [s.name for s in subcommands]:
            raise ManifestException(f"At least one subcommand must have the name {subcommand_basename!r}")

    return Manifest(
        subcommands=subcommands,
        tab_completion_supported=manifest_dict.get("tab_completion_supported", False),
        tab_completion_install_option=manifest_dict.get("tab_completion_install_option"),
        tab_completion_show_option=manifest_dict.get("tab_completion_show_option"),
    )



"""Print version of slices-cli."""

import sys
from importlib.metadata import distributions
from importlib.metadata import version as metadata_version

from . import console, err_console
from .find_subcommands import find_subcommands, warn_if_completion_not_installed
from .get_version import get_executable_version


def print_version():
    """Print the version of slices-cli.

    This also prints the version of the slices subcommands CLIs, and slices clientlibs.
    """

    console.print(f"Slices CLI v{metadata_version('slices-cli')}", highlight=False)
    subcommands = find_subcommands()
    warn_if_completion_not_installed(subcommands)

    core_subcommand = next((subcommand for subcommand in subcommands if subcommand.is_core), None)
    if not core_subcommand:
        err_console.print("[red bold] :x: Slices CLI core is missing.", highlight=False)
        sys.exit(2)

    if core_subcommand.cached_subcommand_version:
        console.print(
            f"   Slices CLI core {core_subcommand.cached_subcommand_version}",
            highlight=False,
        )
    else:
        console.print(
            f"   Slices CLI core {get_executable_version(core_subcommand.executable)}",
            highlight=False,
        )

    for subcommand in subcommands:
        if subcommand.is_core:
            continue

        if subcommand.cached_subcommand_version:
            console.print(
                f"   Slices CLI {subcommand.name} {subcommand.cached_subcommand_version}",
                highlight=False,
            )
        else:
            console.print(
                f"   Slices CLI {subcommand.name} {get_executable_version(subcommand.executable)}",
                highlight=False,
            )

    for dist in distributions():
        if dist.name.startswith("slices-clientlib-"):
            clientlib_name = dist.name[len("slices-clientlib-") :]
            if clientlib_name == "aiis":
                clientlib_name = "ai"
            console.print(f"   Slices clientlib {clientlib_name} v{dist.version}", highlight=False)



"""Find all installed subcommand CLIs."""

import json
import os
import shutil
import sys
from dataclasses import dataclass, field
from importlib.metadata import version as metadata_version
from json import JSONDecodeError
from pathlib import Path
from subprocess import TimeoutExpired, run

from . import console_trace, err_console
from .cache import (
    get_cached_subcommand_info,
    get_cached_subcommand_manifest,
    save_cached_subcommand_info,
    save_cached_subcommand_manifest,
)
from .get_version import get_executable_version
from .manifest import Manifest, ManifestException, process_manifest


@dataclass
class SubCommand:
    """Info about a subcommand CLI."""

    name: str

    executable: Path
    "path of the executable"

    is_ok: bool
    "whether the subcommand is working correctly"

    is_core: bool = False
    "whether the subcommand is a part of slices-cli-core"

    visible_aliases: list[str] = field(default_factory=list)
    "aliases for the subcommand, shown in --help"

    hidden_aliases: list[str] = field(default_factory=list)
    "aliases for the subcommand, not shown in --help or tab completion"

    tab_completion_supported: bool = False
    "whether the subcommand supports command line tab completion"

    tab_completion_install_option: str | list[str] | None = None
    "option to use to install command line tab completion."

    tab_completion_show_option: str | list[str] | None = None
    "option to use to show command line tab completion script."

    short_help: str = "DEFAULT"
    "help oneliner shown when main CLI --help is used"

    cached_subcommand_version: str | None = None
    "version of the subcommand as returned by --version"


def _report_subcommand_manifest_problem(executable: Path, e: Exception):
    """Report a problem with the subcommand manifest."""
    err_console.print(f"[red bold] :x: Error parsing subcommand manifest from {executable}", highlight=False)
    err_console.print(f"[red bold] {getattr(e, 'message', repr(e))}", highlight=False)


def _get_subcommand_manifest(
    executable: Path, *, require_subcommand_with_exe_name: bool = True
) -> tuple[dict | None, Manifest | None]:
    # print(f"_get_subcommand_manifest {executable}")
    try:
        console_trace.print("[bright_black]  executing %s --cli-manifest", executable.absolute())
        manifest_res = run([str(executable.absolute()), "--cli-manifest"], check=False, capture_output=True, timeout=15)  # noqa: S603
        console_trace.print("[bright_black]  returncode: %s", manifest_res.returncode)
        if manifest_res.returncode != 0:
            err_console.print(
                f"[red bold] :x: Error parsing subcommand manifest from {executable} (command failed)", highlight=False
            )
            console_trace.print("[bright_black]  stdout: %s", manifest_res.stdout.decode("utf-8"))
            console_trace.print("[bright_black]  stderr: %s", manifest_res.stderr.decode("utf-8"))
            return None, None

        manifest_data = json.loads(manifest_res.stdout.decode("utf-8"))
        console_trace.print("[bright_black]  loaded manifest data: %s", json.dumps(manifest_data, indent=2))
        return manifest_data, process_manifest(
            executable.name, manifest_data, require_subcommand_with_exe_name=require_subcommand_with_exe_name
        )
    except JSONDecodeError as e:
        _report_subcommand_manifest_problem(executable, e)
        return None, None
    except TimeoutExpired as e:
        _report_subcommand_manifest_problem(executable, e)
        return None, None
    except ManifestException as e:
        _report_subcommand_manifest_problem(executable, e)
        return None, None


# TODO cache subcommand manifest scan
#      use subcommand executable modification time to known when cached manifest is stale


def find_core_subcommands() -> list[SubCommand]:
    """Find slices-cli-core subcommand CLIs."""
    slices_core_which = shutil.which("slices-core")
    if not slices_core_which:
        err_console.print("[red bold] :x: Slices CLI core is not installed.", highlight=False)
        sys.exit(2)

    slices_core_path = Path(slices_core_which)

    if not slices_core_path.is_file() or not os.access(slices_core_path, os.X_OK):
        err_console.print("[red bold] :x: Slices CLI core is incorrectly installed.", highlight=False)
        sys.exit(2)

    # core_version = get_executable_version(slices_core_path)
    core_version = f"v{metadata_version('slices-cli-core')}"
    console_trace.print(f"[bright_black]core_version={core_version}")

    console_trace.print(f"[bright_black]checking core executable: '{slices_core_path}'")

    cached_manifest = get_cached_subcommand_manifest(slices_core_path, "core")
    if not cached_manifest:
        console_trace.print("[bright_black]  no cached core manifest")
        manifest_dict, manifest = _get_subcommand_manifest(slices_core_path, require_subcommand_with_exe_name=False)
        if manifest_dict and manifest:
            console_trace.print("[bright_black]  save core manifest to cache")
            save_cached_subcommand_manifest(slices_core_path, "core", manifest_dict)
        else:
            err_console.print("[red bold] :x: Slices CLI core is incorrectly installed.", highlight=False)
            sys.exit(2)
    else:
        console_trace.print("[bright_black]  using cached core manifest")
        manifest = process_manifest(slices_core_path.name, cached_manifest, require_subcommand_with_exe_name=False)

    if manifest is None:
        err_console.print("[red bold] :x: Slices CLI core is incorrectly installed.", highlight=False)
        sys.exit(2)

    return [
        SubCommand(
            name=subcommand.name,
            executable=slices_core_path,
            tab_completion_supported=manifest.tab_completion_supported,
            tab_completion_install_option=manifest.tab_completion_install_option,
            tab_completion_show_option=manifest.tab_completion_show_option,
            short_help=subcommand.short_help,
            is_ok=True,
            is_core=True,
            cached_subcommand_version=core_version,
            visible_aliases=subcommand.visible_aliases,
            hidden_aliases=subcommand.hidden_aliases,
        )
        for subcommand in manifest.subcommands
    ]

    # fallback
    return [
        SubCommand(
            name="auth",
            executable=slices_core_path,
            tab_completion_supported=True,
            tab_completion_install_option="--install-completion",
            short_help="User authentication",
            is_ok=True,
            is_core=True,
            cached_subcommand_version=core_version,
        ),
        SubCommand(
            name="project",
            executable=slices_core_path,
            tab_completion_supported=True,
            tab_completion_install_option="--install-completion",
            short_help="Projects",
            is_ok=True,
            is_core=True,
            cached_subcommand_version=core_version,
        ),
        SubCommand(
            name="experiment",
            executable=slices_core_path,
            tab_completion_supported=True,
            tab_completion_install_option="--install-completion",
            short_help="Experiments",
            is_ok=True,
            is_core=True,
            cached_subcommand_version=core_version,
        ),
        SubCommand(
            name="pubkey",
            executable=slices_core_path,
            tab_completion_supported=True,
            tab_completion_install_option="--install-completion",
            short_help="SSH key management",
            is_ok=True,
            is_core=True,
            cached_subcommand_version=core_version,
        ),
    ]


# based on from shutil.py
def _get_path() -> list[str] | None:
    path = os.environ.get("PATH", None)
    if path is None:
        try:
            path = os.confstr("CS_PATH")
        except (AttributeError, ValueError):
            # os.confstr() or CS_PATH is not available
            path = os.defpath
    # bpo-35755: Don't use os.defpath if the PATH environment variable
    # is set to an empty string

    # PATH='' doesn't match, whereas PATH=':' looks in the current
    # directory
    if not path:
        return None

    path = os.fsdecode(path)
    return path.split(os.pathsep)


def warn_if_completion_not_installed(subcommands: list[SubCommand]):
    """Warn if no completion is installed for a new subcommand."""

    main_completions_installed = (Path.home() / ".bash_completions" / "slices.sh").is_file() or (
        Path.home() / ".zfunc/_slices"
    ).is_file()
    if not main_completions_installed:
        return

    for subcommand in subcommands:
        if subcommand.is_core:
            continue
        is_installed = get_cached_subcommand_info(
            subcommand.executable, subcommand.name, "completion_installed", invalidate_on_exe_update=False
        )
        if is_installed != "true" and subcommand.tab_completion_supported:
            err_console.print(f"[yellow] :warning: No completion is installed for 'slices-{subcommand.name}'.")
            err_console.print("[yellow] :warning: You need rerun 'slices completion install' to install")
            err_console.print()


def find_subcommands() -> list[SubCommand]:
    """Find all installed subcommand CLIs."""
    core_subcommands = find_core_subcommands()

    known_executables = {"slices-core"}

    res = []

    path = _get_path()
    console_trace.print(f"[bright_black]PATH={path}")
    if not path:
        return core_subcommands

    # find all executables matching slices-*
    for d in path:
        for executable in Path(d).glob("slices-*"):
            if executable.is_file() and os.access(executable, os.X_OK) and executable.name not in known_executables:
                console_trace.print(f"[bright_black]found slices-cli executable: '{executable}'")
                known_executables.add(executable.name)
                name = str(executable.name[len("slices-") :])
                console_trace.print(f"[bright_black]  name={name}")

                cached_manifest = get_cached_subcommand_manifest(executable, name)
                if not cached_manifest:
                    console_trace.print("[bright_black]  no cached manifest")
                    manifest_dict, manifest = _get_subcommand_manifest(executable)
                    if manifest_dict and manifest:
                        console_trace.print("[bright_black]  save manifest to cache")
                        save_cached_subcommand_manifest(executable, name, manifest_dict)
                    else:
                        # save cache failure
                        fake_manifest = {
                            "@type": "slices-cli:subcommand-metadata",
                            "version": "1.0",
                            "tab_completion_supported": False,
                            "short_help": "WORKING INCORRECTLY",
                        }
                        console_trace.print("[bright_black]  save failure to cache")
                        save_cached_subcommand_manifest(executable, name, fake_manifest)
                else:
                    console_trace.print("[bright_black]  using cached manifest")
                    manifest = process_manifest(executable.name, cached_manifest)

                subcommand_version = get_cached_subcommand_info(
                    executable, name, "cli_version", invalidate_on_exe_update=True
                )
                if not subcommand_version:
                    subcommand_version = get_executable_version(executable)
                    save_cached_subcommand_info(executable, name, "cli_version", subcommand_version)

                if manifest is not None:
                    for subcommand in manifest.subcommands:
                        res.append(  # noqa: PERF401
                            SubCommand(
                                name=subcommand.name,
                                executable=executable,
                                tab_completion_supported=manifest.tab_completion_supported,
                                tab_completion_install_option=manifest.tab_completion_install_option,
                                tab_completion_show_option=manifest.tab_completion_show_option,
                                short_help=subcommand.short_help,
                                is_ok=True,
                                cached_subcommand_version=subcommand_version,
                                visible_aliases=subcommand.visible_aliases,
                                hidden_aliases=subcommand.hidden_aliases,
                            )
                        )
                else:
                    res.append(
                        SubCommand(
                            name=name,
                            executable=executable,
                            short_help="WORKING INCORRECTLY",
                            is_ok=False,
                            cached_subcommand_version=subcommand_version,
                        )
                    )

    return core_subcommands + res


def find_subcommand(name: str) -> SubCommand | None:
    """Find a subcommand CLI by name."""
    all = find_subcommands()
    warn_if_completion_not_installed(all)
    for subcommand in all:
        if subcommand.name == name or name in subcommand.visible_aliases or name in subcommand.hidden_aliases:
            return subcommand
    return None



"""Options for slices-cli."""

OPTIONS_WITHOUT_ARGS = [
    "--version",
    "--help",
    "-h",
    "--no-upgrade-check",
    "-v",
    "-vv",
    "-vvv",
]

OPTIONS_WITH_ARGS = []

OPTIONS_WITHOUT_ARGS__HELP = {
    "--version": "Show version and exit.",
    "--help": "Show help and exit.",
    "-h": "Show help and exit.",
    "--no-upgrade-check": "Do not check if new CLI versions are available..",
}

OPTIONS_WITH_ARGS__HELP = {}



"""The main entry point for the slices cli."""

import os
import sys

from slices_cli.completion import install_completion, print_completion_script
from slices_cli.options import OPTIONS_WITH_ARGS, OPTIONS_WITHOUT_ARGS

from . import err_console
from .cache import clear_cache
from .child_exec import exec_subcommand
from .help import print_clear_cache_help, print_completion_help, print_help
from .print_version import print_version
from .upgrade_check import do_upgrade_check


def _next_options(argv: list[str]) -> list[str]:
    res = []
    for arg in argv:
        if arg.startswith("-"):
            res.append(arg)
        else:
            break
    return res


def _has_command(argv: list[str]) -> bool:
    return any(not arg.startswith("-") for arg in argv)


def handle_obsolete_completion():
    """Handle the case where the user is using the old version of command line tab completions."""

    if os.environ.get("_SLICES_COMPLETE", "").startswith("complete_"):
        err_console.print(
            "[red bold] :x: You are using the old version of command line tab completions!", highlight=False
        )
        err_console.print("[red bold] Upgrade with: slices completion install")
        sys.exit(1)


def process_args(argv: list[str]) -> tuple[set[str], list[str], list[str]]:
    """Split the arguments into flags, options with args, and other arguments."""
    argv = list(argv)
    processing_flags = True

    flags = set()
    options_with_arg = []
    other = []

    while len(argv) > 0:
        arg = argv.pop(0)

        if processing_flags:
            if arg in OPTIONS_WITHOUT_ARGS:
                flags.add(arg)
                continue
            if arg in OPTIONS_WITH_ARGS:
                options_with_arg.append(arg)
                options_with_arg.append(argv.pop(0))
                continue

        # unrecognized option, or command: stop processing flags and options
        processing_flags = False
        other.append(arg)
    return flags, options_with_arg, other


def _handle_clear_cache(other: list[str]):
    """Handle the clear-cache command."""
    next_options = _next_options(other)
    non_help_options = [o for o in next_options if o not in ["--help", "-h"]]
    if len(non_help_options) < len(next_options):
        print_clear_cache_help()
        sys.exit(0)

    if _has_command(next_options) or non_help_options:
        err_console.print("[red bold] :x: Unrecognized argument to clear-cache", highlight=False)
        print_clear_cache_help()
        sys.exit(1)

    clear_cache()
    sys.exit(0)


def _handle_completion_command(other: list[str]):
    """Handle the completion command."""
    next_commands = [arg for arg in other if not arg.startswith("-")]
    next_command = next_commands[0] if next_commands else None
    help_options = [o for o in other if o in ["--help", "-h"]]
    non_help_options = [o for o in other if o.startswith("-") and o not in ["--help", "-h"]]

    # wherever there is any --help after completion, show completion help
    if help_options:
        print_completion_help()
        sys.exit(0)

    if non_help_options:
        err_console.print(f"[red bold] :x: Unrecognized option: {non_help_options[0]}", highlight=False)
        print_completion_help()
        sys.exit(1)

    if not next_command:
        err_console.print("[red bold] :x: 'completion' requires a command ('install' or 'show')", highlight=False)
        print_completion_help()
        sys.exit(1)

    if len(next_commands) > 1:
        err_console.print(f"[red bold] :x: Unrecognized argument: {next_commands[1]}", highlight=False)
        print_completion_help()
        sys.exit(1)

    if next_command == "install":
        install_completion()
        sys.exit(0)

    if next_command == "show":
        print_completion_script()
        sys.exit(0)

    if next_command:
        err_console.print(f"[red bold] :x: Unrecognized argument: {next_command}", highlight=False)
        print_completion_help()
        sys.exit(1)


def main():
    """Run the slices cli."""

    handle_obsolete_completion()

    argv = list(sys.argv)

    argv.pop(0)  # pop slices executable

    flags, options_with_arg, other = process_args(argv)

    if "-v" in flags or "-vv" in flags or "-vvv" in flags:
        err_console.print("[red bold]'slices -v' is not supported. Try -v on a subcommand instead.\n")
        print_help()
        sys.exit(1)

    if "--help" in flags or "-h" in flags:
        print_help()
        sys.exit(0)

    if "--version" in flags:
        print_version()
        sys.exit(0)

    # All options with args (options_with_arg) are ignored by this CLI

    while len(other) > 0:
        arg = other.pop(0)
        if arg.startswith("-"):
            # The main slices cli only allows a few specific options.
            # Any other options must be passed to the subcommand.
            err_console.print(f"[red bold] :x: Unrecognized option: {arg}", highlight=False)
            print_help()
            sys.exit(1)

        if arg == "clear-cache":
            _handle_clear_cache(other)

        if arg == "completion":
            _handle_completion_command(other)

        if "--no-upgrade-check" not in flags:
            do_upgrade_check()

        exec_subcommand(arg)

    err_console.print("[red bold] :x: No subcommand specified.")
    print_help()
    sys.exit(1)


if __name__ == "__main__":
    main()



"""Functions for executing subcommand CLIs."""

import os
import shutil
import sys

from . import console, console_trace, err_console
from .find_subcommands import find_subcommand
from .help import print_help


def exec_subcommand(command: str):
    """Execute a subcommand CLI."""

    subcommand = find_subcommand(command)
    if not subcommand:
        err_console.print(f"[red bold] :x: Unrecognized command {command}", highlight=False)
        print_help()
        sys.exit(1)

    # Remove --no-upgrade-check
    argv = [a for a in sys.argv if a != "--no-upgrade-check"]

    console_trace.print(f"Executing {subcommand.executable} with args {argv[1:]}")

    try:
        os.execvp(subcommand.executable, argv)  # noqa: S606
    except:  # noqa: E722
        console.print(f"[red] :x: Could not find {subcommand.executable}. Did you install it correctly?")
        console.print(f"[grey]Debug info: which={shutil.which(subcommand.executable)} PATH={os.environ['PATH']}")
        sys.exit(1)



"""Lazy loading of rich console."""

import os
from typing import Any


class LazyRichConsole:
    """Lazy loading of rich console."""

    def __init__(self, log_path: bool = True, stderr: bool = False):
        """Initialize LazyRichConsole."""
        self.log_path = log_path
        self.stderr = stderr
        self._console = None

    def _lazy_init(self):
        """Initialize _console if not yet done."""

        def make_console(*, log_path: bool = True, stderr: bool = False) -> Any:
            """Make a rich console.

            Includes workaround for special case of python via shell process inside a notebook.
            """

            from rich.console import Console  # noqa: PLC0415  lazy loading of rich

            res = Console(log_path=log_path, stderr=stderr)
            if not res.is_jupyter and os.getenv("JPY_SESSION_NAME") is not None:
                # We are running in a `!` shell process inside a notebook.
                # This is a special case.
                #    - This "terminal" looks interactive, but is not.
                #    - This "terminal" understand color codes, but not terminal control codes to move the cursor etc
                # We force rich to see the terminal as DUMB and non-interactive
                # Also, we make sure we don't revert to default dumb terminal size,
                #    which is very small, so we set fallbacks for LINES and COLUMNS
                os.environ["TERM"] = "DUMB"
                return Console(
                    log_path=log_path,
                    stderr=stderr,
                    force_interactive=False,
                    width=int(os.getenv("COLUMNS") or 140),
                    height=int(os.getenv("LINES") or 30),
                )
            return res

        if self._console is None:
            self._console = make_console(log_path=self.log_path, stderr=self.stderr)

    def print(self, *args, **kwargs):
        """Print to rich console."""
        self._lazy_init()
        self._console.print(*args, **kwargs)  # pyright: ignore[reportOptionalMemberAccess]


class DummyConsole:
    """Console that does nothing."""

    def print(self, *args, **kwargs):
        """Do nothing."""
